part of '../../public.dart';

abstract class ApiController<OutputType> {
  static late final Dio apiDio;
  ApiController({
    required this.dio,
    required this.name,
    required this.fromMap,
  });
  final Dio dio;
  final String name;
  final OutputType Function(Map<String, dynamic> data) fromMap;
  String pathBuild([String? methodPath]) => "/$name/${methodPath ?? ""}";

  Future<List<OutputType>> get({
    Map<String, dynamic>? queryParameters,
    Options? options,
    CancelToken? cancelToken,
    void Function(int total, int finish)? onReceiveProgress,
  }) async {
    var response = await dio.get<String>(
      pathBuild(),
      queryParameters: queryParameters,
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
    );

    if (response.statusCode == 200) {
      return response.data?.toListMap().map(fromMap).toList() ?? [];
    } else {
      throw response.statusMessage ?? response.data ?? "Error";
    }
  }

  Future<OutputType?> getById(
    Object id, {
    Options? options,
    CancelToken? cancelToken,
    void Function(int total, int finish)? onReceiveProgress,
  }) async {
    final response = await dio.get<String>(
      pathBuild(),
      queryParameters: {"id": id},
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
    );
    if (response.statusCode == 200) {
      final map = response.data?.toMap();
      if (map == null) return null;
      return fromMap(map);
    } else {
      throw response.statusMessage ?? response.data ?? "Error";
    }
  }

  Future<OutputType?> post(
    Map<String, dynamic> data, {
    Options? options,
    CancelToken? cancelToken,
    void Function(int total, int finish)? onReceiveProgress,
    void Function(int total, int finish)? onSendProgress,
  }) async {
    var response = await dio.post<String>(
      pathBuild(),
      data: data.toJson(),
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
      onSendProgress: onSendProgress,
    );

    if (response.statusCode == 200) {
      final map = response.data?.toMap();
      if (map == null) return null;
      return fromMap(map);
    } else {
      throw response.statusMessage ?? response.data ?? "Error";
    }
  }

  Future<OutputType?> put(
    Map<String, Object> primaryKeys,
    Map<String, dynamic> data, {
    Options? options,
    CancelToken? cancelToken,
    void Function(int total, int finish)? onReceiveProgress,
    void Function(int total, int finish)? onSendProgress,
  }) async {
    final response = await dio.put<String>(
      pathBuild(),
      data: data.toJson(),
      queryParameters: primaryKeys,
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
      onSendProgress: onSendProgress,
    );

    if (response.statusCode == 200) {
      final map = response.data?.toMap();
      if (map == null) return null;
      return fromMap(map);
    } else {
      throw response.statusMessage ?? response.data ?? "Error";
    }
  }

  Future<OutputType?> delete(
    Map<String, Object> primaryKeys, {
    Options? options,
    CancelToken? cancelToken,
  }) async {
    var response = await dio.delete<dynamic>(
      pathBuild(),
      queryParameters: primaryKeys,
      options: options,
      cancelToken: cancelToken,
    );

    if (response.statusCode == 200) {
      final map = response.data?.toMap();
      if (map == null) return null;
      return fromMap(map);
    } else {
      throw response.statusMessage ?? response.data ?? "Error";
    }
  }
}

class ModelApiController<OutputType extends Model,
        DbSetType extends ModelDbSet<OutputType>>
    extends ApiController<OutputType> {
  ModelApiController({
    required super.dio,
    required super.name,
    required super.fromMap,
    required this.dbSet,
    this.rows = const {},
  }) : super();

  final DbSetType dbSet;
  final Set<OutputType> rows;

  @override
  Future<List<OutputType>> get({
    Map<String, dynamic>? queryParameters,
    Options? options,
    CancelToken? cancelToken,
    void Function(int total, int finish)? onReceiveProgress,
    bool caching = false,
  }) async {
    final response = await super.get(
      queryParameters: queryParameters,
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
    );

    if (caching) {
      dbSet.insertAll(response).then((value) => debugPrint("$value"));
    }
    rows.addAll(response);
    return response;
  }

  @override
  Future<OutputType?> getById(
    Object id, {
    Options? options,
    CancelToken? cancelToken,
    void Function(int total, int finish)? onReceiveProgress,
    bool caching = false,
  }) async {
    final response = await super.getById(
      id,
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
    );
    if (response == null) return null;

    if (caching) dbSet.insertOne(response);
    rows.add(response);
    return response;
  }

  @override
  Future<OutputType?> post(
    Map<String, dynamic> data, {
    Options? options,
    CancelToken? cancelToken,
    void Function(int total, int finish)? onReceiveProgress,
    void Function(int total, int finish)? onSendProgress,
    bool caching = false,
  }) async {
    final response = await super.post(
      data,
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
      onSendProgress: onSendProgress,
    );
    if (response == null) return null;

    if (caching) dbSet.insertOne(response);
    rows.add(response);
    return response;
  }

  @override
  Future<OutputType?> put(
    Map<String, Object> primaryKeys,
    Map<String, dynamic> data, {
    Options? options,
    CancelToken? cancelToken,
    void Function(int total, int finish)? onReceiveProgress,
    void Function(int total, int finish)? onSendProgress,
    bool caching = false,
  }) async {
    final response = await super.put(
      primaryKeys,
      data,
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
      onSendProgress: onSendProgress,
    );
    if (response == null) return null;

    if (caching) dbSet.insertOne(response);
    rows.add(response);
    return response;
  }

  @override
  Future<OutputType?> delete(
    Map<String, Object> primaryKeys, {
    Options? options,
    CancelToken? cancelToken,
    bool caching = false,
  }) async {
    final response = await super.delete(
      primaryKeys,
      options: options,
      cancelToken: cancelToken,
    );
    if (response == null) return null;

    if (caching) dbSet.insertOne(response);
    rows.add(response);
    return response;
  }
}
